
.hrepr-True, .hrepr-False, .hrepr-NoneType, .hrepr-int, .hrepr-float {
	font-weight: bold;
}

.hrepr-int, .hrepr-float {
	color: #88f;
}

.hrepr-str {
	color: #666;
	white-space: pre;
}

.hrepr-str:before {
	content: "“";
	color: #888;
}

.hrepr-str:after {
	content: "”";
	color: #888;
}

.hrepr-True {
	color: green;
}

.hrepr-False {
	color: red;
}

.hrepr-NoneType {
	color: grey;
}

.hrepr-list, .hrepr-tuple, .hrepr-set, .hrepr-frozenset, .hrepr-dict {
	display: inline-block;
}

.hrepr-dict {
	border: 1px solid grey;
	border-radius: 5px;
}

.hrepr-list > *,
.hrepr-tuple > *,
.hrepr-set > *,
.hrepr-frozenset > *,
.hrepr-dict > * {
	vertical-align: middle;
}
